@@ -544,19 +544,37 @@ static PyMethodDef tp_methods[] =
 static PyObject *
 tp_new (PyTypeObject *type, PyObject *args, PyObject *)
 {
-  // parse input args
-  PyObject *arg_reader = nullptr;
+  // Parse input args
+  mobius::io::reader arg_reader;
 
-  if (!PyArg_ParseTuple (args, "O", &arg_reader))
-    return nullptr;
+  try
+    {
+      arg_reader = PyMobius_IO_Reader_as_cpp (mobius::py::get_arg (args, 0));
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
 
-  // create object
-  decoder_mfc_decoder_o *self = (decoder_mfc_decoder_o *) type->tp_alloc (type, 0);
+  // Create Python object
+  decoder_mfc_decoder_o *ret = reinterpret_cast <decoder_mfc_decoder_o *> (type->tp_alloc (type, 0));
 
-  if (self)
-    self->obj = new mobius::decoder::mfc_decoder (*((io_reader_o *) arg_reader)->obj);
+  if (ret)
+    {
+      try
+        {
+          ret->obj = new mobius::decoder::mfc_decoder (arg_reader);
+        }
+      catch (const std::exception& e)
+        {
+          Py_DECREF (ret);
+          PyErr_SetString (PyExc_Exception, e.what ());
+          ret = nullptr;
+        }
+    }
 
-  return (PyObject *) self;
+  return reinterpret_cast <PyObject *> (ret);
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
