@@ -1,6 +1,6 @@
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 // Mobius Forensic Toolkit
-// Copyright (C) 2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018,2019 Eduardo Aguiar
+// Copyright (C) 2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018,2019,2020 Eduardo Aguiar
 //
 // This program is free software; you can redistribute it and/or modify it
 // under the terms of the GNU General Public License as published by the
@@ -30,6 +30,8 @@
 #include "password.h"
 #include "password_hash.h"
 #include "profile.h"
+#include "text_search.h"
+#include "pod/data.h"
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 //! \brief Create <i>item</i> Python object from C++ object
@@ -101,6 +103,59 @@ tp_getter_category (model_item_o *self)
   return ret;
 }
 
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+//! \brief <i>metadata</i> attribute getter
+//! \param self object
+//! \return <i>metadata</i> attribute
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+static PyObject *
+tp_getter_metadata (model_item_o *self)
+{
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyMobius_Pod_Data_from_cpp (self->obj->get_metadata ());
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  return ret;
+}
+
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+//! \brief <i>metadata</i> Attribute setter
+//! \param self Object
+//! \param value New value
+//! \return 0 on success, -1 on error
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+static int
+tp_setter_metadata (model_item_o *self, PyObject *value, void *)
+{
+  // If it is a delete operation, then returns with error
+  if (value == nullptr)
+    {
+      PyErr_SetString (PyExc_TypeError, "cannot delete 'metadata' attribute");
+      return -1;
+    }
+
+  // Set attribute
+  try
+    {
+      self->obj->set_metadata (PyMobius_Pod_Data_as_cpp (value));
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+      return -1;
+    }
+
+  // Return success
+  return 0;
+}
+
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 //! \brief <i>case</i> attribute getter
 //! \param self object
@@ -142,6 +197,13 @@ static PyGetSetDef tp_getset[] =
     (char *) "category",
     NULL
   },
+  {
+    (char *) "metadata",
+    (getter) tp_getter_metadata,
+    (setter) tp_setter_metadata,
+    (char *) "metadata",
+    NULL
+  },
   {
     (char *) "case",
     (getter) tp_getter_case,
@@ -268,7 +330,6 @@ tp_f_new_child (model_item_o *self, PyObject *args)
 static PyObject *
 tp_f_remove (model_item_o *self, PyObject *)
 {
-  // execute C++ function
   try
     {
       self->obj->remove ();
@@ -279,9 +340,7 @@ tp_f_remove (model_item_o *self, PyObject *)
       return nullptr;
     }
 
-  // return None
-  Py_INCREF (Py_None);
-  return Py_None;
+  return mobius::py::pynone ();
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -311,8 +370,7 @@ tp_f_move (model_item_o *self, PyObject *args)
     }
 
   // return None
-  Py_INCREF (Py_None);
-  return Py_None;
+  return mobius::py::pynone ();
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -325,10 +383,17 @@ static PyObject *
 tp_f_has_attribute (model_item_o *self, PyObject *args)
 {
   // parse input args
-  const char * arg_id;
+  std::string arg_id;
 
-  if (!PyArg_ParseTuple (args, "s", &arg_id))
-    return nullptr;
+  try
+    {
+      arg_id = mobius::py::get_arg_as_std_string (args, 0);
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
 
   // execute C++ function
   PyObject *ret = nullptr;
@@ -356,10 +421,17 @@ static PyObject *
 tp_f_get_attribute (model_item_o *self, PyObject *args)
 {
   // parse input args
-  const char * arg_id;
+  std::string arg_id;
 
-  if (!PyArg_ParseTuple (args, "s", &arg_id))
-    return nullptr;
+  try
+    {
+      arg_id = mobius::py::get_arg_as_std_string (args, 0);
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
 
   // execute C++ function
   PyObject *ret = nullptr;
@@ -386,11 +458,19 @@ static PyObject *
 tp_f_set_attribute (model_item_o *self, PyObject *args)
 {
   // parse input args
-  const char * arg_id;
-  const char * arg_value;
+  std::string arg_id;
+  std::string arg_value;
 
-  if (!PyArg_ParseTuple (args, "ss", &arg_id, &arg_value))
-    return nullptr;
+  try
+    {
+      arg_id = mobius::py::get_arg_as_std_string (args, 0);
+      arg_value = mobius::py::get_arg_as_std_string (args, 1);
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
 
   // execute C++ function
   try
@@ -403,9 +483,7 @@ tp_f_set_attribute (model_item_o *self, PyObject *args)
       return nullptr;
     }
 
-  // return None
-  Py_INCREF (Py_None);
-  return Py_None;
+  return mobius::py::pynone ();
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -417,10 +495,17 @@ static PyObject *
 tp_f_remove_attribute (model_item_o *self, PyObject *args)
 {
   // parse input args
-  const char * arg_id;
+  std::string arg_id;
 
-  if (!PyArg_ParseTuple (args, "s", &arg_id))
-    return nullptr;
+  try
+    {
+      arg_id = mobius::py::get_arg_as_std_string (args, 0);
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
 
   // execute C++ function
   try
@@ -434,8 +519,7 @@ tp_f_remove_attribute (model_item_o *self, PyObject *args)
     }
 
   // return None
-  Py_INCREF (Py_None);
-  return Py_None;
+  return mobius::py::pynone ();
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -473,7 +557,6 @@ tp_f_get_attributes (model_item_o *self, PyObject *)
 static PyObject *
 tp_f_expand_masks (model_item_o *self, PyObject *)
 {
-  // execute C++ function
   try
     {
       self->obj->expand_masks ();
@@ -484,9 +567,7 @@ tp_f_expand_masks (model_item_o *self, PyObject *)
       return nullptr;
     }
 
-  // return None
-  Py_INCREF (Py_None);
-  return Py_None;
+  return mobius::py::pynone ();
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -499,10 +580,17 @@ static PyObject *
 tp_f_get_data_path (model_item_o *self, PyObject *args)
 {
   // parse input args
-  const char * arg_rpath;
+  std::string arg_rpath;
 
-  if (!PyArg_ParseTuple (args, "s", &arg_rpath))
-    return nullptr;
+  try
+    {
+      arg_rpath = mobius::py::get_arg_as_std_string (args, 0);
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
 
   // execute C++ function
   PyObject *ret = nullptr;
@@ -530,10 +618,17 @@ static PyObject *
 tp_f_create_data_path (model_item_o *self, PyObject *args)
 {
   // parse input args
-  const char * arg_rpath;
+  std::string arg_rpath;
 
-  if (!PyArg_ParseTuple (args, "s", &arg_rpath))
-    return nullptr;
+  try
+    {
+      arg_rpath = mobius::py::get_arg_as_std_string (args, 0);
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
 
   // execute C++ function
   PyObject *ret = nullptr;
@@ -561,10 +656,17 @@ static PyObject *
 tp_f_has_ant (model_item_o *self, PyObject *args)
 {
   // parse input args
-  const char * arg_id;
+  std::string arg_id;
 
-  if (!PyArg_ParseTuple (args, "s", &arg_id))
-    return nullptr;
+  try
+    {
+      arg_id = mobius::py::get_arg_as_std_string (args, 0);
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
 
   // execute C++ function
   PyObject *ret = nullptr;
@@ -591,12 +693,21 @@ static PyObject *
 tp_f_set_ant (model_item_o *self, PyObject *args)
 {
   // parse input args
-  const char * arg_id;
-  const char * arg_name;
-  const char * arg_version;
+  std::string arg_id;
+  std::string arg_name;
+  std::string arg_version;
 
-  if (!PyArg_ParseTuple (args, "sss", &arg_id, &arg_name, &arg_version))
-    return nullptr;
+  try
+    {
+      arg_id = mobius::py::get_arg_as_std_string (args, 0);
+      arg_name = mobius::py::get_arg_as_std_string (args, 1);
+      arg_version = mobius::py::get_arg_as_std_string (args, 2);
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
 
   // execute C++ function
   try
@@ -610,8 +721,7 @@ tp_f_set_ant (model_item_o *self, PyObject *args)
     }
 
   // return None
-  Py_INCREF (Py_None);
-  return Py_None;
+  return mobius::py::pynone ();
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -623,10 +733,17 @@ static PyObject *
 tp_f_reset_ant (model_item_o *self, PyObject *args)
 {
   // parse input args
-  const char * arg_id;
+  std::string arg_id;
 
-  if (!PyArg_ParseTuple (args, "s", &arg_id))
-    return nullptr;
+  try
+    {
+      arg_id = mobius::py::get_arg_as_std_string (args, 0);
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
 
   // execute C++ function
   try
@@ -639,9 +756,7 @@ tp_f_reset_ant (model_item_o *self, PyObject *args)
       return nullptr;
     }
 
-  // return None
-  Py_INCREF (Py_None);
-  return Py_None;
+  return mobius::py::pynone ();
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -652,7 +767,6 @@ tp_f_reset_ant (model_item_o *self, PyObject *args)
 static PyObject *
 tp_f_remove_ants (model_item_o *self, PyObject *)
 {
-  // execute C++ function
   try
     {
       self->obj->remove_ants ();
@@ -663,9 +777,7 @@ tp_f_remove_ants (model_item_o *self, PyObject *)
       return nullptr;
     }
 
-  // return None
-  Py_INCREF (Py_None);
-  return Py_None;
+  return mobius::py::pynone ();
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -703,12 +815,21 @@ static PyObject *
 tp_f_new_password (model_item_o *self, PyObject *args)
 {
   // parse input args
-  const char * arg_type;
-  const char * arg_value;
-  const char * arg_description;
+  std::string arg_type;
+  std::string arg_value;
+  std::string arg_description;
 
-  if (!PyArg_ParseTuple (args, "sss", &arg_type, &arg_value, &arg_description))
-    return nullptr;
+  try
+    {
+      arg_type = mobius::py::get_arg_as_std_string (args, 0);
+      arg_value = mobius::py::get_arg_as_std_string (args, 1);
+      arg_description = mobius::py::get_arg_as_std_string (args, 2);
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
 
   // execute C++ function
   PyObject *ret = nullptr;
@@ -761,7 +882,6 @@ tp_f_get_passwords (model_item_o *self, PyObject *)
 static PyObject *
 tp_f_remove_passwords (model_item_o *self, PyObject *)
 {
-  // execute C++ function
   try
     {
       self->obj->remove_passwords ();
@@ -772,9 +892,7 @@ tp_f_remove_passwords (model_item_o *self, PyObject *)
       return nullptr;
     }
 
-  // return None
-  Py_INCREF (Py_None);
-  return Py_None;
+  return mobius::py::pynone ();
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -787,12 +905,21 @@ static PyObject *
 tp_f_new_password_hash (model_item_o *self, PyObject *args)
 {
   // parse input args
-  const char * arg_type;
-  const char * arg_value;
-  const char * arg_description;
+  std::string arg_type;
+  std::string arg_value;
+  std::string arg_description;
 
-  if (!PyArg_ParseTuple (args, "sss", &arg_type, &arg_value, &arg_description))
-    return nullptr;
+  try
+    {
+      arg_type = mobius::py::get_arg_as_std_string (args, 0);
+      arg_value = mobius::py::get_arg_as_std_string (args, 1);
+      arg_description = mobius::py::get_arg_as_std_string (args, 2);
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
 
   // execute C++ function
   PyObject *ret = nullptr;
@@ -844,7 +971,6 @@ tp_f_get_password_hashes (model_item_o *self, PyObject *)
 static PyObject *
 tp_f_remove_password_hashes (model_item_o *self, PyObject *)
 {
-  // execute C++ function
   try
     {
       self->obj->remove_password_hashes ();
@@ -855,9 +981,7 @@ tp_f_remove_password_hashes (model_item_o *self, PyObject *)
       return nullptr;
     }
 
-  // return None
-  Py_INCREF (Py_None);
-  return Py_None;
+  return mobius::py::pynone ();
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -936,7 +1060,6 @@ tp_f_get_cookies (model_item_o *self, PyObject *)
 static PyObject *
 tp_f_remove_cookies (model_item_o *self, PyObject *)
 {
-  // execute C++ function
   try
     {
       self->obj->remove_cookies ();
@@ -947,9 +1070,7 @@ tp_f_remove_cookies (model_item_o *self, PyObject *)
       return nullptr;
     }
 
-  // return None
-  Py_INCREF (Py_None);
-  return Py_None;
+  return mobius::py::pynone ();
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -962,11 +1083,19 @@ static PyObject *
 tp_f_new_profile (model_item_o *self, PyObject *args)
 {
   // parse input args
-  const char * arg_app_id;
-  const char * arg_path;
+  std::string arg_app_id;
+  std::string arg_path;
 
-  if (!PyArg_ParseTuple (args, "ss", &arg_app_id, &arg_path))
-    return nullptr;
+  try
+    {
+      arg_app_id = mobius::py::get_arg_as_std_string (args, 0);
+      arg_path = mobius::py::get_arg_as_std_string (args, 1);
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
 
   // execute C++ function
   PyObject *ret = nullptr;
@@ -1019,10 +1148,17 @@ static PyObject *
 tp_f_remove_profiles_by_app_id (model_item_o *self, PyObject *args)
 {
   // parse input args
-  const char * arg_app_id;
+  std::string arg_app_id;
 
-  if (!PyArg_ParseTuple (args, "s", &arg_app_id))
-    return nullptr;
+  try
+    {
+      arg_app_id = mobius::py::get_arg_as_std_string (args, 0);
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
 
   // execute C++ function
   try
@@ -1035,8 +1171,96 @@ tp_f_remove_profiles_by_app_id (model_item_o *self, PyObject *args)
     }
 
   // return None
-  Py_INCREF (Py_None);
-  return Py_None;
+  return mobius::py::pynone ();
+}
+
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+//! \brief <i>new_text_search</i> method implementation
+//! \param self Object
+//! \param args Argument list
+//! \return Text search
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+static PyObject *
+tp_f_new_text_search (model_item_o *self, PyObject *args)
+{
+  // parse input args
+  mobius::datetime::datetime arg_timestamp;
+  std::string arg_type;
+  std::string arg_text;
+
+  try
+    {
+      arg_timestamp = mobius::py::get_arg_as_datetime (args, 0);
+      arg_type = mobius::py::get_arg_as_std_string (args, 1);
+      arg_text = mobius::py::get_arg_as_std_string (args, 2);
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_TypeError, e.what ());
+      return nullptr;
+    }
+
+  // Execute C++ function
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = PyMobius_Model_Text_Search_from_cpp (self->obj->new_text_search (arg_timestamp, arg_type, arg_text));
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+
+  // Return value
+  return ret;
+}
+
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+//! \brief <i>get_text_searches</i> method implementation
+//! \param self Object
+//! \param args Argument list
+//! \return None
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+static PyObject *
+tp_f_get_text_searches (model_item_o *self, PyObject *)
+{
+  PyObject *ret = nullptr;
+
+  try
+    {
+      ret = mobius::py::pylist_from_cpp_container (
+              self->obj->get_text_searches (),
+              PyMobius_Model_Text_Search_from_cpp
+            );
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+    }
+    
+  return ret;
+}
+
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+//! \brief <i>remove_text_searches</i> method implementation
+//! \param self Object
+//! \param args Argument list
+// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
+static PyObject *
+tp_f_remove_text_searches (model_item_o *self, PyObject *)
+{
+  try
+    {
+      self->obj->remove_text_searches ();
+    }
+  catch (const std::exception& e)
+    {
+      PyErr_SetString (PyExc_Exception, e.what ());
+      return nullptr;
+    }
+
+  return mobius::py::pynone ();
 }
 
 // =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
@@ -1230,6 +1454,24 @@ static PyMethodDef tp_methods[] =
     METH_VARARGS,
     "Remove profiles by application ID"
   },
+  {
+    (char *) "new_text_search",
+    (PyCFunction) tp_f_new_text_search,
+    METH_VARARGS,
+    "Create new text search object"
+  },
+  {
+    (char *) "get_text_searches",
+    (PyCFunction) tp_f_get_text_searches,
+    METH_VARARGS,
+    "Get text searches"
+  },
+  {
+    (char *) "remove_text_searches",
+    (PyCFunction) tp_f_remove_text_searches,
+    METH_VARARGS,
+    "Remove text searches"
+  },
   {NULL, NULL, 0, NULL} // sentinel
 };
 
